mod daemon;
mod listenbrainz;
mod models;
mod playlist;
mod subsonic;
mod track;

use crate::daemon::update_playlists_daemon;
use crate::models::{CreateTrackedPlaylist, PartialTrackedPlaylist, TrackedPlaylist};
use crate::subsonic::{Subsonic, SubsonicBuilder};
use config::Config;
use rocket::serde::json::{json, Json};
use rocket::{get, post, routes, serde::json::Value as JsonValue, State};
use serde::Serialize;
use sqlx::postgres::PgPool;
use std::str::FromStr;
use std::sync::{Arc, OnceLock};
use uuid::Uuid;

pub static CONFIG: OnceLock<Config> = OnceLock::new();

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    CONFIG
        .set(
            Config::builder()
                .add_source(config::File::with_name("/etc/playlistd").required(false))
                .add_source(config::Environment::default().separator("_"))
                .build()
                .unwrap(),
        )
        .unwrap();

    let database = PgPool::connect(
        CONFIG
            .get()
            .unwrap()
            .get::<String>("database.url")?
            .as_str(),
    )
    .await?;
    sqlx::migrate!().run(&database).await?;

    let media_client = Arc::new(
        SubsonicBuilder::new()
            .base(CONFIG.get().unwrap().get::<String>("subsonic.base")?)
            .username(CONFIG.get().unwrap().get::<String>("subsonic.username")?)
            .password(CONFIG.get().unwrap().get::<String>("subsonic.password")?)
            .build()?,
    );

    let _media_client = media_client.clone();
    tokio::spawn(async move {
        update_playlists_daemon(_media_client).await.unwrap();
    });

    rocket::build()
        .manage(media_client)
        .manage(database)
        .mount(
            "/",
            routes![
                create_tracking_playlist,
                add_tracking_playlist,
                all_playlists,
            ],
        )
        .launch()
        .await?;

    Ok(())
}

#[derive(Serialize)]
struct JsonError {
    error: String,
}

impl From<sqlx::Error> for JsonError {
    fn from(value: sqlx::Error) -> Self {
        Self {
            error: value.to_string(),
        }
    }
}

impl From<uuid::Error> for JsonError {
    fn from(value: uuid::Error) -> Self {
        Self {
            error: value.to_string(),
        }
    }
}

#[derive(Serialize)]
enum Response<T> {
    Success(T),
    Error(JsonError),
}

#[post("/create", data = "<create_playlist>")]
async fn create_tracking_playlist(
    media_client: &State<Arc<Subsonic>>,
    pool: &State<PgPool>,
    create_playlist: Json<CreateTrackedPlaylist>,
) -> Json<Response<TrackedPlaylist>> {
    match media_client
        .create_playlist(create_playlist.playlist_name.clone())
        .await
    {
        Ok(created_playlist) => match Uuid::from_str(&created_playlist.id) {
            Ok(playlist_id) => {
                let partial = PartialTrackedPlaylist {
                    playlist_id,
                    playlist_size: create_playlist.playlist_size,
                    tracking_user: create_playlist.tracking_user.clone(),
                    tracking_type: create_playlist.tracking_type.clone(),
                };

                match partial.record(pool.inner()).await {
                    Ok(playlist) => Json(Response::Success(playlist)),
                    Err(e) => Json(Response::Error(e.into())),
                }
            }

            Err(e) => Json(Response::Error(e.into())),
        },

        Err(_) => Json(Response::Error(JsonError {
            error: "Couldn't create Subsonic playlist".to_string(),
        })),
    }
}

#[post("/add", data = "<playlist>")]
async fn add_tracking_playlist(
    pool: &State<PgPool>,
    playlist: Json<PartialTrackedPlaylist>,
) -> Json<Response<TrackedPlaylist>> {
    match playlist.record(pool.inner()).await {
        Ok(playlist) => Json(Response::Success(playlist)),
        Err(e) => Json(Response::Error(e.into())),
    }
}

#[get("/")]
async fn all_playlists(pool: &State<PgPool>) -> JsonValue {
    match TrackedPlaylist::all(pool.inner()).await {
        Ok(playlists) => json!(playlists),
        Err(_) => json!([]),
    }
}
